/*
 * Copyright 2006-2007 Bas Leijdekkers
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.intellij.plugins.lineage.actions;

import com.intellij.ide.util.NavigationItemListCellRenderer;
import com.intellij.ide.util.PsiClassListCellRenderer;
import com.intellij.openapi.actionSystem.*;
import com.intellij.openapi.editor.CaretModel;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.editor.SelectionModel;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.popup.JBPopup;
import com.intellij.openapi.ui.popup.PopupChooserBuilder;
import com.intellij.openapi.wm.WindowManager;
import com.intellij.openapi.wm.StatusBar;
import com.intellij.openapi.keymap.KeymapUtil;
import com.intellij.psi.*;
import com.intellij.psi.tree.IElementType;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.ui.ListSpeedSearch;
import com.intellij.codeInsight.hint.HintManager;
import net.intellij.plugins.lineage.tools.HighlightTool;

import javax.swing.AbstractListModel;
import javax.swing.JList;
import javax.swing.ListCellRenderer;
import javax.swing.ListSelectionModel;
import java.io.Serializable;
import java.util.*;

/** @noinspection ComponentNotRegistered*/
public class HighlightUsagesAction extends OverridingAction
{
	protected boolean performAction(AnActionEvent event)
	{
		final Editor editor = event.getData(DataKeys.EDITOR);
		if (editor == null)
		{
			return false;
		}
		final SelectionModel selectionModel = editor.getSelectionModel();
		if (selectionModel.hasSelection())
		{
			return false;
		}
		final Project project = editor.getProject();
		if (project == null)
		{
			return false;
		}
		final PsiDocumentManager documentManager = PsiDocumentManager.getInstance(project);
		final PsiFile file = documentManager.getPsiFile(editor.getDocument());
		if (file == null)
		{
			return false;
		}
		final CaretModel caretModel = editor.getCaretModel();
		final int offset = caretModel.getOffset();
		final PsiElement element = file.findElementAt(offset);
		if (element == null)
		{
			return false;
		}
		if (!(element instanceof PsiJavaToken))
		{
			return false;
		}
		final PsiJavaToken token = (PsiJavaToken) element;
		final IElementType tokenType = token.getTokenType();
		return JavaTokenType.IMPORT_KEYWORD.equals(tokenType) &&
				performHighlightImportedTypeAction(element, editor);
	}

	private boolean performHighlightImportedTypeAction(final PsiElement element,
	                                                          Editor editor)
	{
		final PsiElement parent = element.getParent();
		if (!(parent instanceof PsiImportStatementBase))
		{
			return false;
		}
		final PsiImportStatementBase importStatement = (PsiImportStatementBase) parent;
		final PsiJavaCodeReferenceElement importReference = importStatement.getImportReference();
		if (importReference == null)
		{
			return false;
		}
		final PsiJavaFile javaFile =
				PsiTreeUtil.getParentOfType(importReference, PsiJavaFile.class);
		if (javaFile == null)
		{
			return false;
		}
		final PsiElement importedElement = importReference.resolve();
		if (importedElement == null)
		{
			return false;
		}
		final boolean onDemand = importStatement.isOnDemand();
		final boolean importStatic = importStatement instanceof PsiImportStaticStatement;
		final ReferenceCollector collector =
				new ReferenceCollector(importedElement, onDemand, importStatic);
		javaFile.accept(collector);
		final Map<PsiMember, List<PsiElement>> classReferenceListMap =
				collector.getClassReferenceListMap();
		if (classReferenceListMap.isEmpty())
		{
			HintManager.getInstance().showInformationHint(editor, "No usages of import found");
			return true;
		}
		if (classReferenceListMap.size() <= 1)
		{
			// single import
			final Collection<List<PsiElement>> lists = classReferenceListMap.values();
			final List<PsiElement> importReferences = lists.toArray(new List[1])[0];
			importReferences.add(element);
			final boolean highlighted = HighlightTool.highlightElements(importReferences);
			setStatusBar(element.getProject(), importReferences.size() - 1, highlighted);
			return true;
		}
		final List<PsiMember> importedMembers = new ArrayList(classReferenceListMap.keySet());
		final ListCellRenderer renderer;
		if (importStatic)
		{
			renderer = new NavigationItemListCellRenderer();
		}
		else
		{
			renderer = new PsiClassListCellRenderer();
		}
		Collections.sort(importedMembers, new PsiMemberComparator());
		final JList list = new JList(new LineageListModel(importedMembers));
		list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		list.setCellRenderer(renderer);
		new LineageListSpeedSearch(list);
		final Runnable callback = new Runnable()
		{
			public void run()
			{
				// on demand import
				final int index= list.getSelectedIndex();
				if (index > 0)
				{
					final PsiMember importedClass = importedMembers.get(index - 1);
					final List<PsiElement> referenceList = classReferenceListMap.get(importedClass);
					referenceList.add(element);
					final boolean highlighted = HighlightTool.highlightElements(referenceList);
					setStatusBar(element.getProject(), referenceList.size() - 1, highlighted);
				}
				else if (index == 0)
				{
					final Collection<List<PsiElement>> referenceLists =
							classReferenceListMap.values();
					final List<PsiElement> elementsToHighlight = new ArrayList();
					elementsToHighlight.add(element);
					for (List<PsiElement> referenceList : referenceLists)
					{
						elementsToHighlight.addAll(referenceList);
					}
					final boolean highlighted =
							HighlightTool.highlightElements(elementsToHighlight);
					setStatusBar(element.getProject(), elementsToHighlight.size() - 1, highlighted);
				}
			}
		};
		final PopupChooserBuilder builder = new PopupChooserBuilder(list);
		if (importStatic)
		{
			builder.setTitle("Choose Imported Members to Highlight");
		}
		else
		{
			builder.setTitle("Choose Imported Classes to Highlight");
		}
		builder.setItemChoosenCallback(callback);
		final JBPopup popup = builder.createPopup();
		popup.showInBestPositionFor(editor);
		return true;
	}

	private void setStatusBar(Project project,
			int count, boolean highlighted)
	{
		final WindowManager windowManager = WindowManager.getInstance();
		final StatusBar statusBar = windowManager.getStatusBar(project);
		if (highlighted)
		{
			if (count > 1)
			{
				statusBar.setInfo(count + " usages of import found (press " + getShortcutText() +
						" again to remove the highlighting, Escape to remove all highlighting)");
			}
			else
			{
				statusBar.setInfo(count + " usage of import found (press " + getShortcutText() +
						" again to remove the highlighting, Escape to remove all highlighting)");
			}
		}
		else
		{
			statusBar.setInfo("");
		}
	}

	private String getShortcutText()
	{
		final ShortcutSet shortcutSet = getShortcutSet();
		final Shortcut shortcut = shortcutSet.getShortcuts()[0];
		return KeymapUtil.getShortcutText(shortcut);
	}

	private static class LineageListModel extends AbstractListModel
	{
		private final List<Object> classes;

		LineageListModel(List<? extends PsiMember> classes)
		{
			this.classes = new ArrayList();
			this.classes.add(0, "All listed");
			this.classes.addAll(classes);
		}

		public int getSize()
		{
			return classes.size();
		}

		public Object getElementAt(int index)
		{
			return classes.get(index);
		}
	}

	private static class PsiMemberComparator implements Comparator<PsiMember>, Serializable
	{
		public int compare(PsiMember member1, PsiMember member2)
		{
			final String name1 = member1.getName();
			if (name1 == null)
			{
				return -1;
			}
			final String name2 = member2.getName();
			if (name2 == null)
			{
				return 1;
			}
			final int i = name1.compareTo(name2);
			if (i != 0)
			{
				return i;
			}
			final PsiJavaFile file1 = (PsiJavaFile)member1.getContainingFile();
			final PsiJavaFile file2 = (PsiJavaFile)member2.getContainingFile();
			final String packageName1 = file1.getPackageName();
			final String packageName2 = file2.getPackageName();
			return packageName1.compareTo(packageName2);
		}
	}

	private static class LineageListSpeedSearch extends ListSpeedSearch
	{
		private LineageListSpeedSearch(JList list)
		{
			super(list);
		}

		protected String getElementText(Object object)
		{
			if (!(object instanceof PsiMember))
			{
				return object.toString();
			}
			final PsiMember member = (PsiMember)object;
			final PsiJavaFile file = (PsiJavaFile)member.getContainingFile();
			return member.getName() + ' ' + file.getPackageName();
		}
	}
}
