/*
 * Copyright 2006-2007 Bas Leijdekkers
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.intellij.plugins.lineage.actions;

import com.intellij.openapi.actionSystem.AnAction;
import com.intellij.openapi.actionSystem.AnActionEvent;
import com.intellij.openapi.actionSystem.ActionManager;

public abstract class OverridingAction extends AnAction
{
	private AnAction overriddenAction;

	protected OverridingAction()
	{
	}

	public final void actionPerformed(AnActionEvent event)
	{
		if (!performAction(event))
		{
			overriddenAction.actionPerformed(event);
		}
	}

	/**
	 * Override this method if you want the overriding action to have different state and
	 * presentation change behavior.
	 */
	public boolean doUpdate(AnActionEvent event)
	{
		return false;
	}

	/**
	 * Register this OverridingAction to override the action with the id specified.
	 */
	public void override(String overriddenActionId)
	{
		final ActionManager actionManager = ActionManager.getInstance();
		overriddenAction = actionManager.getAction(overriddenActionId);
		if (overriddenAction == null)
		{
			throw new IllegalArgumentException("unknown action: " + overriddenActionId);
		}
		actionManager.unregisterAction(overriddenActionId);
		copyFrom(overriddenAction);
		setInjectedContext(overriddenAction.isInInjectedContext());
		setEnabledInModalContext(overriddenAction.isEnabledInModalContext());
		setDefaultIcon(overriddenAction.isDefaultIcon());
		actionManager.registerAction(overriddenActionId, this);
	}

	/**
	 * @return true, if the overriding action could be performed, false if not and the overridden
	 * action should still be performed.
	 */
	protected abstract boolean performAction(AnActionEvent event);

	/**
	 * Restore an overridden action back to the overridden action.
	 * @param overriddenActionId  the id of the overridden action
	 */
	public static void restore(String overriddenActionId)
	{
		final ActionManager actionManager = ActionManager.getInstance();
		final AnAction anAction = actionManager.getAction(overriddenActionId);
		if (anAction == null)
		{
			throw new IllegalArgumentException(overriddenActionId + " action not registered");
		}
		if (!(anAction instanceof OverridingAction))
		{
			throw new IllegalArgumentException(overriddenActionId +
					" action is not an overriding action");
		}
		final OverridingAction overridingAction = (OverridingAction) anAction;
		actionManager.unregisterAction(overriddenActionId);
		actionManager.registerAction(overriddenActionId, overridingAction.overriddenAction);
	}

	/**
	 * Calls the update method of the overridden action so it has the same state and presentation
	 * changes.
	 * @see #doUpdate(AnActionEvent)
	 */
	public final void update(AnActionEvent event)
	{
		if (!doUpdate(event))
		{
			overriddenAction.update(event);
		}
	}
}
