/*
 * Copyright 2007 Bas Leijdekkers
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.intellij.plugins.lineage.actions;

import com.intellij.openapi.actionSystem.*;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.openapi.ui.Messages;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.command.CommandProcessor;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.application.Application;
import com.intellij.openapi.roots.ProjectRootManager;
import com.intellij.openapi.roots.ProjectFileIndex;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.psi.*;
import com.intellij.psi.search.GlobalSearchScope;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.util.IncorrectOperationException;
import org.jetbrains.annotations.Nullable;

import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;
import java.io.File;
import java.util.*;

/** @noinspection ComponentNotRegistered*/
public class PasteFilesAction extends OverridingAction
{
	private List<File> fileList = null;

	public boolean doUpdate(AnActionEvent event)
	{
		fileList = getClipboardFiles();
		if (!isAvailable(event, fileList))
		{
			return false;
		}
		final Presentation presentation = event.getPresentation();
		presentation.setEnabled(true);
		if (fileList.size() == 1)
		{
			final File file = fileList.get(0);
			if (file.isDirectory())
			{
				presentation.setText("Paste Directory");
			}
			else
			{
				presentation.setText("Paste File");
			}
		}
		else
		{
			boolean directories = true;
			for (File file : fileList)
			{
				if (!file.isDirectory())
				{
					directories = false;
					break;
				}
			}
			if (directories)
			{
				presentation.setText("Paste Directories");
			}
			else
			{
				presentation.setText("Paste Files");
			}
		}
		return true;
	}

	private static boolean isAvailable(AnActionEvent event, List<File> clipboardFiles)
	{
		if (clipboardFiles == null || clipboardFiles.isEmpty())
		{
			return false;
		}
		final PsiElement element = event.getData(DataKeys.PSI_ELEMENT);
		final PsiDirectory directory;
		if (element instanceof PsiDirectory)
		{
			directory = (PsiDirectory)element;
		}
		else if (element != null)
		{
			final PsiFile containingFile = element.getContainingFile();
			if (containingFile == null)
			{
				return false;
			}
			directory = containingFile.getContainingDirectory();
		}
		else
		{
			return false;
		}
		if (directory == null ||
				directory.getPackage() == null && !directory.isSourceRoot())
		{
			return false;
		}
		final ProjectRootManager rootManager =
				ProjectRootManager.getInstance(event.getData(DataKeys.PROJECT));
		final ProjectFileIndex fileIndex = rootManager.getFileIndex();
		final VirtualFile virtualFile = directory.getVirtualFile();
		return !(fileIndex.isInLibraryClasses(virtualFile) ||
				fileIndex.isInLibrarySource(virtualFile));
	}

	@Nullable
	private static List<File> getClipboardFiles()
	{
		final Toolkit toolkit = Toolkit.getDefaultToolkit();
		final Clipboard clipboard = toolkit.getSystemClipboard();
		final Transferable transferable = clipboard.getContents(null);
		if (transferable == null)
		{
			return null;
		}
		try
		{
			if (transferable.isDataFlavorSupported(DataFlavor.javaFileListFlavor))
			{
				return (List<File>)transferable.getTransferData(DataFlavor.javaFileListFlavor);
			}
		}
		catch (UnsupportedFlavorException e)
		{
			throw new RuntimeException(e);
		}
		catch (IOException e)
		{
			throw new RuntimeException(e);
		}
		return null;
	}

	public boolean performAction(AnActionEvent event)
	{
		final PsiElement element = event.getData(DataKeys.PSI_ELEMENT);
		final PsiDirectory directory;
		if (element instanceof PsiDirectory)
		{
			directory = (PsiDirectory)element;
		}
		else
		{
			final PsiFile containingFile = element.getContainingFile();
			directory = containingFile.getContainingDirectory();
		}
		final String packagePrefix;
		final PsiPackage aPackage = directory.getPackage();
		if (aPackage != null)
		{
			packagePrefix = aPackage.getQualifiedName();
		}
		else
		{
			packagePrefix = "";
		}
		final PsiManager manager = directory.getManager();
//		directory.c
		final PsiElementFactory factory = manager.getElementFactory();
		/*final PsiJavaFile javaFile =
				(PsiJavaFile)factory.createFileFromText(DUMMY_FILE_NAME, fileList);
		final PsiElement[] children = javaFile.getChildren();
		int startOffset = 0;
		String currentPackageName = null;
		String commonPrefix = null;
		final Map<String, List<PsiFile>> packageNames2FileList = new HashMap();
		final Map<String, List<String>> packageNames2ClassNameList = new HashMap();
		String firstFileName = null;
		String firstPackagName = null;
		for (PsiElement child : children)
		{
			if (child instanceof PsiPackageStatement)
			{
				// can only happen once at the beginning of the file.
				final PsiPackageStatement packageStatement =
						(PsiPackageStatement) child;
				currentPackageName = packageStatement.getPackageName();
			}
			else if (child instanceof PsiClass)
			{
				final PsiClass aClass = (PsiClass) child;
				final int endOffset = aClass.getTextRange().getEndOffset();
				final String sourceText =
						fileList.substring(startOffset, endOffset);
				final PsiFile newFile =
						factory.createFileFromText(DUMMY_FILE_NAME, sourceText);
				startOffset = endOffset;
				if (newFile instanceof PsiJavaFile)
				{
					final PsiJavaFile newJavaFile = (PsiJavaFile)newFile;
					final String previousPackageName = currentPackageName;
					currentPackageName = newJavaFile.getPackageName();
					if (currentPackageName.length() == 0)
					{
						currentPackageName = previousPackageName;
					}
					else
					{
						if (commonPrefix == null)
						{
							if (previousPackageName != null)
							{
								commonPrefix = StringUtil.commonPrefix(
										currentPackageName, previousPackageName);
							}
						}
						else
						{
							commonPrefix = StringUtil.commonPrefix(commonPrefix,
									currentPackageName);
						}
					}
				}
				final String className = aClass.getName();
				try
				{
					newFile.setName(className + ".java");
				}
				catch (IncorrectOperationException e)
				{
					throw new RuntimeException(e);
				}
				if (currentPackageName == null)
				{
					currentPackageName = "";
				}
				List<PsiFile> fileList = packageNames2FileList.get(currentPackageName);
				List<String> classNameList = packageNames2ClassNameList.get(currentPackageName);
				if (fileList == null)
				{
					fileList = new ArrayList();
					classNameList = new ArrayList();
					packageNames2FileList.put(currentPackageName, fileList);
					packageNames2ClassNameList.put(currentPackageName, classNameList);
					if (firstFileName == null)
					{
						firstFileName = className + ".java";
						firstPackagName = currentPackageName;
					}
				}
				fileList.add(newFile);
				classNameList.add(className);
			}
			else if (child instanceof PsiErrorElement)
			{
				final PsiErrorElement errorElement = (PsiErrorElement) child;
				final String description = errorElement.getErrorDescription();
				if (!description.equals("'class' or 'interface' expected"))
				{
					continue;
				}
				if (!errorElement.getText().startsWith("package"))
				{
					continue;
				}
				final PsiElement previousSibling =
						PsiTreeUtil.skipSiblingsBackward(errorElement,
								PsiWhiteSpace.class);
				if (previousSibling instanceof PsiComment)
				{
					startOffset = previousSibling.getTextOffset();
				}
				else
				{
					startOffset = errorElement.getTextOffset();
				}
			}*/

		/*}

		if (packagePrefix.length() == 0 || commonPrefix == null)
		{
			commonPrefix = "";
		}
		final int stripPrefixCharacters = commonPrefix.length();
		if (!checkCreateClasses(packageNames2ClassNameList, directory, stripPrefixCharacters))
		{
			return false;
		}
		final CommandProcessor commandProcessor = CommandProcessor.getInstance();
		final Project project = event.getData(DataKeys.PROJECT);
		final String finalFirstFileName = firstFileName;
		final String finalFirstPackagName = firstPackagName;
		commandProcessor.executeCommand(project, new Runnable()
		{
			public void run()
			{
				final Application application = ApplicationManager.getApplication();
				application.runWriteAction(
						new Runnable()
						{
							public void run()
							{
								try
								{
									for (String packageName : packageNames2FileList.keySet())
									{
										final String strippedPackageName;
										if (packagePrefix.length() == 0)
										{
											strippedPackageName = packageName;
										}
										else
										{
											strippedPackageName =
													packageName.substring(stripPrefixCharacters);
										}
										PsiDirectory targetDirectory = directory;
										if (strippedPackageName.length() > 0)
										{
											final String[] subdirectoryNames =
													strippedPackageName.split("\\.");
											for (String subdirectoryName : subdirectoryNames)
											{
												final PsiDirectory subdirectory =
														targetDirectory.findSubdirectory(
																subdirectoryName);
												if (subdirectory == null)
												{
													targetDirectory =
															targetDirectory.createSubdirectory(
																	subdirectoryName);
												}
												else
												{
													targetDirectory = subdirectory;
												}
											}
										}
										final List<PsiFile> fileList =
												packageNames2FileList.get(packageName);
										final PsiPackage targetPackage =
												targetDirectory.getPackage();
										final String targetPackageName;
										if (targetPackage != null)
										{
											targetPackageName = targetPackage.getQualifiedName();
										}
										else
										{
											targetPackageName = "";
										}
										final PsiPackageStatement packageStatement;
										if (targetPackageName.length() > 0)
										{
											packageStatement =
													factory.createPackageStatement(targetPackageName);
										}
										else
										{
											packageStatement = null;
										}
										for (PsiFile file : fileList)
										{
											final PsiJavaFile result =
													(PsiJavaFile)targetDirectory.add(file);
											final PsiPackageStatement oldPackageStatement =
													result.getPackageStatement();
											if (oldPackageStatement != null)
											{
												if (packageStatement != null)
												{
													oldPackageStatement.replace(packageStatement);
												}
												else
												{
													oldPackageStatement.delete();
												}
											}
											else if (packageStatement != null)
											{
												result.add(packageStatement);
											}
											if (file.getName().equals(finalFirstFileName) &&
													packageName.equals(finalFirstPackagName))
											{
												result.navigate(true);
											}
										}
									}
								}
								catch (IncorrectOperationException e)
								{
									throw new RuntimeException(e);
								}
							}
						}
				);
			}
		}, "Paste Class", "Paste Class");*/
		return true;
	}

	private static boolean checkCreateClasses(Map<String, List<String>> packageNames2ClassNameList,
	                                       PsiDirectory directory, int stripPrefixCharacters)
	{
		final PsiManager manager = directory.getManager();
		forloop:
		for (String packageName : packageNames2ClassNameList.keySet())
		{
			final String strippedPackageName;
			if (directory.isSourceRoot())
			{
				strippedPackageName = packageName;
			}
			else
			{
				strippedPackageName = packageName.substring(stripPrefixCharacters);
			}
			PsiDirectory targetDirectory = directory;
			if (strippedPackageName.length() > 0)
			{
				final String[] subdirectoryNames = strippedPackageName.split("\\.");
				for (String subdirectoryName : subdirectoryNames)
				{
					targetDirectory = targetDirectory.findSubdirectory(subdirectoryName);
					if (targetDirectory == null)
					{
						continue forloop;
					}
				}
			}
			final List<String> classNameList =
					packageNames2ClassNameList.get(packageName);
			for (String className : classNameList)
			{
				final PsiPackage directoryPackage = targetDirectory.getPackage();
				final String name;
				if (directoryPackage != null)
				{
					name = directoryPackage.getQualifiedName();
				}
				else
				{
					name = "";
				}
				final GlobalSearchScope scope =
						GlobalSearchScope.directoryScope(targetDirectory, false);
				final String qualifiedName;
				if (name.length() > 0)
				{
					qualifiedName = name + '.' + className;
				}
				else
				{
					qualifiedName = className;
				}
				final PsiClass foundClass = manager.findClass(qualifiedName, scope);
				if (foundClass != null)
				{
					final Project project = directory.getProject();
					final String path =
							targetDirectory.getVirtualFile().getPresentableUrl();
					final String type;
					if (foundClass.isInterface())
					{
						type = "an interface";
					}
					else if (foundClass.isAnnotationType())
					{
						type = "an @interface";
					}
					else if (foundClass.isEnum())
					{
						type = "an enum";
					}
					else
					{
						type = "a class";
					}
					Messages.showErrorDialog(project,
							"Directory " + path +
									"\nalready contains " + type + " named '" +
									className + '\'', "Error");
					return false;
				}
			}
		}
		return true;
	}
}