/*
 * Copyright 2006-2007 Bas Leijdekkers
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.intellij.plugins.lineage.actions;

import com.intellij.psi.*;
import com.intellij.psi.util.InheritanceUtil;
import com.intellij.psi.util.PsiTreeUtil;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

class ReferenceCollector extends PsiRecursiveElementVisitor
{
	private final Map<PsiMember, List<PsiElement>> classReferenceListMap = new HashMap();
	private final PsiElement importTarget;
	private final boolean onDemand;
	private final boolean importStatic;

	ReferenceCollector(@NotNull PsiElement importTarget, boolean onDemand, boolean importStatic)
	{
		this.importTarget = importTarget;
		this.onDemand = onDemand;
		this.importStatic = importStatic;
	}

	public void visitReferenceElement(PsiJavaCodeReferenceElement reference)
	{
		super.visitReferenceElement(reference);
		if (isFullyQualifiedReference(reference))
		{
			return;
		}
		PsiElement parent = reference.getParent();
		if (parent instanceof PsiImportStatementBase)
		{
			return;
		}
		while (parent instanceof PsiJavaCodeReferenceElement)
		{
			parent = parent.getParent();
			if (parent instanceof PsiImportStatementBase)
			{
				return;
			}
		}
		if (importStatic)
		{
			checkStaticImportReference(reference);
		}
		else
		{
			checkImportReference(reference);
		}
	}

	private void checkStaticImportReference(PsiJavaCodeReferenceElement reference)
	{
		if (reference.isQualified())
		{
			return;
		}
		final PsiElement target = reference.resolve();
		if (!(target instanceof PsiMethod) &&
				!(target instanceof PsiClass) &&
				!(target instanceof PsiField))
		{
			return;
		}
		final PsiMember member = (PsiMember) target;
		if (importTarget instanceof PsiMethod)
		{
			if (member.equals(importTarget))
			{
				addReference(member, reference);
			}
		}
		else if (importTarget instanceof PsiClass)
		{
			final PsiClass importClass = (PsiClass)importTarget;
			if (onDemand)
			{
				final PsiClass containingClass = member.getContainingClass();
				if (InheritanceUtil.isInheritorOrSelf(importClass, containingClass, true))
				{
					addReference(member, reference);
				}
			}
			else
			{
				if (importTarget.equals(member))
				{
					addReference(member, reference);
				}
			}
		}
	}

	private void checkImportReference(PsiJavaCodeReferenceElement reference)
	{
		final PsiElement element = reference.resolve();
		if (!(element instanceof PsiClass))
		{
			return;
		}
		final PsiClass referencedClass = (PsiClass) element;
		if (importTarget instanceof PsiPackage)
		{
			if (referencedClass.getContainingClass() != null)
			{
				return;
			}
			final PsiFile file = referencedClass.getContainingFile();
			if (!(file instanceof PsiJavaFile))
			{
				return;
			}
			final PsiJavaFile javaFile = (PsiJavaFile) file;
			final PsiPackage aPackage = (PsiPackage) importTarget;
			final String packageName = aPackage.getQualifiedName();
			final String filePackage = javaFile.getPackageName();
			if (filePackage.equals(packageName))
			{
				addReference(referencedClass, reference);
			}
		}
		else if (importTarget instanceof PsiClass)
		{
			final PsiClass aClass = (PsiClass) importTarget;
			final String name = aClass.getQualifiedName();
			if (name == null)
			{
				return;
			}
			if (!onDemand)
			{
				if (name.equals(referencedClass.getQualifiedName()))
				{
					addReference(referencedClass, reference);
				}
			}
			else
			{
				final PsiClass containingClass = referencedClass.getContainingClass();
				if (containingClass == null)
				{
					return;
				}
				if (name.equals(containingClass.getQualifiedName()))
				{
					addReference(referencedClass, reference);
				}
			}
		}
	}

	private void addReference(PsiMember referencedMember, PsiJavaCodeReferenceElement reference)
	{
		List<PsiElement> referenceList =
				classReferenceListMap.get(referencedMember);
		if (referenceList == null)
		{
			referenceList = new ArrayList();
			classReferenceListMap.put(referencedMember, referenceList);
		}
		referenceList.add(reference.getReferenceNameElement());
	}

	public Map<PsiMember, List<PsiElement>> getClassReferenceListMap()
	{
		return classReferenceListMap;
	}

	public static boolean isFullyQualifiedReference(PsiJavaCodeReferenceElement reference)
	{
		if (!reference.isQualified())
		{
			return false;
		}
		final PsiElement directParent = reference.getParent();
		if (directParent instanceof PsiMethodCallExpression ||
				directParent instanceof PsiAssignmentExpression ||
				directParent instanceof PsiVariable)
		{
			return false;
		}
		final PsiElement parent = PsiTreeUtil.getParentOfType(reference,
				PsiImportStatementBase.class, PsiPackageStatement.class,
				PsiCodeFragment.class);
		if (parent != null)
		{
			return false;
		}
		final PsiElement target = reference.resolve();
		if(!(target instanceof PsiClass))
		{
			return false;
		}
		final PsiClass aClass = (PsiClass) target;
		final String fqName = aClass.getQualifiedName();
		if (fqName == null)
		{
			return false;
		}
		final String text = stripAngleBrackets(reference.getText());
		return text.equals(fqName);
	}

	private static String stripAngleBrackets(String string)
	{
		final int index = string.indexOf('<');
		if (index == -1)
		{
			return string;
		}
		return string.substring(0, index);
	}
}