/*
 * Copyright 2006-2007 Bas Leijdekkers
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.intellij.plugins.lineage.actions;

import com.intellij.openapi.actionSystem.AnAction;
import com.intellij.openapi.actionSystem.AnActionEvent;
import com.intellij.openapi.actionSystem.DataKeys;
import com.intellij.openapi.actionSystem.Presentation;
import com.intellij.openapi.command.CommandProcessor;
import com.intellij.psi.*;
import com.intellij.ide.IdeView;
import net.intellij.plugins.lineage.ui.SubClassDialog;
import net.intellij.plugins.lineage.tools.SubclassType;
import net.intellij.plugins.lineage.tools.SubclassTools;

public class SubclassAction extends AnAction
{

	public SubclassAction()
	{
		super("Subclass/Subinterface", "Extend Class or Interface", null);
	}

	public void update(AnActionEvent event)
	{
		final Presentation presentation = event.getPresentation();
		final PsiElement element = event.getData(DataKeys.PSI_ELEMENT);
		if (element instanceof PsiClass)
		{
			final PsiClass aClass = (PsiClass)element;
			if (!aClass.isEnum())
			{
				presentation.setVisible(true);
				presentation.setEnabled(!aClass.hasModifierProperty(PsiModifier.FINAL));
				if (aClass.isInterface())
				{
					presentation.setText("Extend interface");
				}
				else
				{
					presentation.setText("Extend class");
				}
				return;
			}
		}
		presentation.setVisible(false);
		presentation.setEnabled(false);
	}

	public void actionPerformed(AnActionEvent event)
	{
		final PsiElement element = event.getData(DataKeys.PSI_ELEMENT);
		if (!(element instanceof PsiClass)) {
			return;
		}
		final PsiClass superClass = (PsiClass)element;
		final PsiFile file = event.getData(DataKeys.PSI_FILE);
		final PsiElement context;
		if (file == null) {
			context = superClass;
		} else {
			context = file;
		}
		final CommandProcessor commandProcessor = CommandProcessor.getInstance();
		final String text;
		final SubclassType subclassType;
		if (superClass.isInterface())
		{
			text = "Extend Interface";
			subclassType = SubclassType.INTERFACE;
		}
		else if (superClass.hasModifierProperty(PsiModifier.ABSTRACT))
		{
			text = "Extend Abstract Class";
			subclassType = SubclassType.CLASS;
		}
		else
		{
			text = "Extend Class";
			subclassType = SubclassType.CLASS;
		}
		final IdeView view = event.getData(DataKeys.IDE_VIEW);
		commandProcessor.executeCommand(superClass.getProject(), new Runnable()
		{
			public void run()
			{
				final PsiManager manager = superClass.getManager();
				final PsiElementFactory factory = manager.getElementFactory();
				final PsiClassType superClassType = factory.createType(superClass);
				final String className = "My" + superClass.getName();
				final SubClassDialog dialog = new SubClassDialog(className, superClassType,
						context, subclassType, false, false);
				dialog.show();
				if (!dialog.isOK())
				{
					return;
				}
				final PsiDirectory targetDirectory = dialog.getTargetDirectory();
				final PsiClass newClass = SubclassTools
						.createTopLevelSubclass(dialog.getSubclassName(), superClassType,
								targetDirectory, subclassType, dialog.getVisibility(),
								dialog.shouldMakeAbstract());
				if (view != null && newClass != null)
				{
					view.selectElement(targetDirectory);
					final PsiJavaToken brace = newClass.getLBrace();
					view.selectElement(brace);
				}
			}
		}, text, text);
	}
}