/*
 * Copyright 2006-2007 Bas Leijdekkers
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.intellij.plugins.lineage.intentions;

import com.intellij.openapi.application.Application;
import com.intellij.openapi.application.ApplicationInfo;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.command.CommandProcessor;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.project.Project;
import com.intellij.psi.*;
import com.intellij.psi.tree.IElementType;
import com.intellij.psi.codeStyle.CodeStyleManager;
import com.intellij.psi.search.GlobalSearchScope;
import com.intellij.psi.util.PsiElementFilter;
import com.intellij.util.IncorrectOperationException;
import net.intellij.plugins.lineage.tools.ExpectedTypeTool;
import net.intellij.plugins.lineage.ui.SubClassDialog;
import net.intellij.plugins.lineage.tools.SubclassTools;
import net.intellij.plugins.lineage.tools.SubclassType;
import org.jetbrains.annotations.NotNull;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

public class LineageCreateClassFromUsageFix extends LineageIntention
{
	private final SubclassType subclassType;
	private final boolean createInnerClass;

	public LineageCreateClassFromUsageFix(SubclassType subclassType,
	                                      boolean createInnerClass)
	{
		this.subclassType = subclassType;
		this.createInnerClass = createInnerClass;
	}

	protected PsiElementFilter getElementFilter()
	{
		return new CreateClassFilter(subclassType);
	}

	protected String getTextForElement(PsiElement element)
	{
		final String elementText = element.getText();
		if (createInnerClass)
		{
			return "Create Inner Class '" + elementText + '\'';
		}
		else if (subclassType == SubclassType.CLASS)
		{
			return "Create Class '" + elementText + '\'';
		}
		else if (subclassType == SubclassType.INTERFACE)
		{
			return "Create Interface '" + elementText + '\'';
		}
		else if (subclassType == SubclassType.ENUM)
		{
			return "Create Enum '" + elementText + '\'';
		}
		else if (subclassType == SubclassType.ANNOTATION_TYPE)
		{
			return "Create @interface '" + elementText + '\'';
		}
		throw new InternalError();
	}

	@NotNull
	public String getFamilyName()
	{
		if (createInnerClass)
		{
			return "Create Inner Class from Usage";
		}
		else if (subclassType == SubclassType.CLASS)
		{
			return "Create Class from Usage";
		}
		else if (subclassType == SubclassType.ENUM)
		{
			return "Create Enum from Usage";
		}
		else if (subclassType == SubclassType.INTERFACE)
		{
			return "Create Interface from Usage";
		}
		else if (subclassType == SubclassType.ANNOTATION_TYPE)
		{
			return "Create Annotation Type from Usage";
		}
		throw new InternalError();
	}

	public boolean startInWriteAction()
	{
		return false;
	}

	private static void setupGenericParameters(
			PsiJavaCodeReferenceElement referenceElement,
			PsiClass targetClass)
			throws IncorrectOperationException
	{
		final int parameterCount = referenceElement.getTypeParameters().length;
		if (parameterCount == 0)
		{
			return;
		}
		final PsiManager manager = referenceElement.getManager();
		final PsiElementFactory factory = manager.getElementFactory();
		final PsiTypeParameterList parameterList =
				targetClass.getTypeParameterList();
		if (parameterList == null)
		{
			return;
		}
		parameterList.add(factory.createTypeParameterFromText("T", null));
		for (int i = 2; i <= parameterCount; i++) {
			final PsiTypeParameter typeParameter =
					factory.createTypeParameterFromText("T" + (i - 1), null);
			parameterList.add(typeParameter);
		}
	}

	public void invoke(@NotNull final Project project, final Editor editor,
	                   final PsiFile file)
			throws IncorrectOperationException
	{
		final PsiElement element = findMatchingElement(file, editor);
		if (element == null)
		{
			return;
		}
		final PsiElement parent = element.getParent();
		if (!(parent instanceof PsiJavaCodeReferenceElement))
		{
			return;
		}
		final PsiJavaCodeReferenceElement referenceElement =
				(PsiJavaCodeReferenceElement)parent;
		final PsiElement grandParent = referenceElement.getParent();
		final boolean createConstructorFromCall;
		PsiType type;
		if (grandParent instanceof PsiNewExpression)
		{
			final PsiNewExpression newExpression =
					(PsiNewExpression)grandParent;
			final PsiExpressionList argumentList =
					newExpression.getArgumentList();
			if (argumentList != null)
			{
				final PsiExpression[] arguments = argumentList.getExpressions();
				createConstructorFromCall = arguments.length > 0;
			}
			else
			{
				createConstructorFromCall = false;
			}
			type = ExpectedTypeTool.findExpectedType(newExpression, false);
		}
		else if (grandParent instanceof PsiTypeElement)
		{
			createConstructorFromCall = false;
			type = null;
		}
		else if (grandParent instanceof PsiReferenceList)
		{
			final PsiReferenceList referenceList = (PsiReferenceList)grandParent;
			final PsiElement firstChild = referenceList.getFirstChild();
			if (firstChild instanceof PsiKeyword)
			{
				final PsiKeyword keyword = (PsiKeyword)firstChild;
				final IElementType tokenType = keyword.getTokenType();
				if (JavaTokenType.THROWS_KEYWORD.equals(tokenType))
				{
					final PsiManager manager = element.getManager();
					final GlobalSearchScope resolveScope =
							element.getResolveScope();
					final PsiElementFactory factory =
							manager.getElementFactory();
					type = factory.createTypeByFQClassName("java.lang.Exception",
							resolveScope);
				}
				else
				{
					type = null;
				}
			}
			else {
				type = null;
			}
			createConstructorFromCall = false;
		}
		else if (grandParent instanceof PsiAnnotation)
		{
			createConstructorFromCall = false;
			type = null;
		}
		else
		{
			return;
		}
		while (type instanceof PsiArrayType)
		{
			final PsiArrayType arrayType = (PsiArrayType)type;
			type = arrayType.getComponentType();
		}
		final PsiClassType superClassType;
		if (type == null)
		{
			final PsiManager manager = PsiManager.getInstance(project);
			final GlobalSearchScope scope = grandParent.getResolveScope();
			superClassType = PsiType.getJavaLangObject(manager, scope);
		}
		else if (type instanceof PsiClassType)
		{
			superClassType = (PsiClassType)type;
		}
		else
		{
			return;
		}
		final String text = getTextForElement(element);
		final String subclassName = element.getText();
		final SubClassDialog dialog =
				new SubClassDialog(subclassName, superClassType, element,
						subclassType, createInnerClass, true);
		dialog.setCheckVisibilityFromContext(true);
		dialog.setTitle(getFamilyName());
		dialog.show();
		if (!dialog.isOK())
		{
			return;
		}
		final CommandProcessor commandProcessor = CommandProcessor.getInstance();
		commandProcessor.executeCommand(project, new Runnable()
		{
			public void run()
			{
				final PsiClass newClass;
				final String subclassName = dialog.getSubclassName();
				final String visibility = dialog.getVisibility();
				final boolean makeAbstract = dialog.shouldMakeAbstract();
				if (createInnerClass)
				{
					newClass = SubclassTools.createNestedSubclass(subclassName, superClassType,
							dialog.getTargetClass(), subclassType, visibility,
							dialog.shouldMakeStatic(), makeAbstract);
				}
				else
				{
					newClass = SubclassTools.createTopLevelSubclass(subclassName, superClassType,
							dialog.getTargetDirectory(),
							subclassType, visibility, makeAbstract);
				}
				if (newClass == null)
				{
					return;
				}
				final Application application = ApplicationManager.getApplication();
				application.runWriteAction(new Runnable()
				{

					public void run()
					{
						try
						{
							if (subclassType != SubclassType.ENUM &&
									subclassType != SubclassType.ANNOTATION_TYPE) {
								setupGenericParameters(referenceElement, newClass);
							}
							final PsiManager manager = newClass.getManager();
							final CodeStyleManager styleManager = manager.getCodeStyleManager();
							final PsiElement formattedClass = styleManager.reformat(newClass);
							referenceElement.bindToElement(formattedClass);
						}
						catch (IncorrectOperationException e)
						{
							throw new RuntimeException(e);
						}
					}
				});
				newClass.navigate(true);
			}
		}, text, text);
		if (createConstructorFromCall)
		{
			final Application application = ApplicationManager.getApplication();
			application.invokeLater(new Runnable()
			{
				public void run()
				{
					final CommandProcessor commandProcessor = CommandProcessor.getInstance();
					commandProcessor.executeCommand(project, new Runnable()
					{
						public void run()
						{
							application.runWriteAction(new Runnable()
							{
								public void run()
								{
									final PsiNewExpression newExpression = (PsiNewExpression)grandParent;
									final ApplicationInfo info = ApplicationInfo.getInstance();
									final int build = Integer.parseInt(info.getBuildNumber());
									final Class<?> createConstructorFromCallClass;
									try
								 	{
										if(build < 6696)
										{
											createConstructorFromCallClass =
													Class.forName("com.intellij.codeInsight.daemon.impl.quickfix." +
															"CreateConstructorFromCallAction");
										}
										else
										{

											createConstructorFromCallClass =
													Class.forName("com.intellij.codeInsight.daemon.impl.quickfix." +
															"CreateConstructorFromCallFix");
										}
									}
									catch (ClassNotFoundException ignore)
									{
										// probably api changed, silently continue
										return;
									}
									try
									{
										final Constructor<?> constructor =
												createConstructorFromCallClass
														.getConstructor(PsiConstructorCall.class);
										final Object fixObject =
												constructor.newInstance(newExpression);
										final Method method =
												createConstructorFromCallClass.getMethod("invoke",
														Project.class, Editor.class, PsiFile.class);
										method.invoke(fixObject, project, editor, file);
									}
									catch (NoSuchMethodException ignore)
									{
										// probably api changed, silently continue
									}
									catch (InvocationTargetException ignore)
									{
										// probably api changed, silently continue
									}
									catch (IllegalAccessException ignore)
									{
										// probably api changed, silently continue
									}
									catch (InstantiationException ignore)
									{
										// probably api changed, silently continue
									}
								}
							});
						}
					}, "Create Constructor from Usage", "Create Constructor from Usage");
				}
			});
		}
	}
}