/*
 * Copyright 2006-2007 Bas Leijdekkers
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.intellij.plugins.lineage.intentions;

import com.intellij.openapi.command.CommandProcessor;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.project.Project;
import com.intellij.psi.*;
import com.intellij.psi.util.PsiElementFilter;
import com.intellij.util.IncorrectOperationException;
import net.intellij.plugins.lineage.ui.SubClassDialog;
import net.intellij.plugins.lineage.tools.SubclassType;
import net.intellij.plugins.lineage.tools.SubclassTools;
import org.jetbrains.annotations.NotNull;

public class SubClassIntention extends LineageIntention
{
	final boolean innerClass;
	final SubclassType subclassType;

	public SubClassIntention(boolean innerClass, SubclassType subclassType)
	{
		this.innerClass = innerClass;
		this.subclassType = subclassType;
	}

	protected PsiElementFilter getElementFilter()
	{
		return new SubClassFilter();
	}

	@NotNull
	public String getFamilyName()
	{
		if (subclassType != SubclassType.INTERFACE)
		{
			if (innerClass)
			{
				return "Implement Interface with Inner";
			}
			else
			{
				return "Implement Interface";
			}
		}
		else
		{
			if (innerClass)
			{
				return "Extend Classes and Interfaces with Inner";
			}
			else
			{
				return "Extend Classes and Interfaces";
			}
		}
	}

	protected String getTextForElement(PsiElement element)
	{
		final PsiElement parent = element.getParent();
		final PsiClass aClass;
		if (parent instanceof PsiClass)
		{
			aClass = (PsiClass)parent;
		}
		else
		{
			final PsiJavaCodeReferenceElement reference = (PsiJavaCodeReferenceElement)parent;
			aClass = (PsiClass)reference.resolve();
		}
		if (aClass == null)
		{
			return "";
		}
		if (aClass.isInterface())
		{
			if (subclassType != SubclassType.INTERFACE)
			{
				if (innerClass)
				{
					return "Implement Interface with Inner Class";
				}
				else
				{
					return "Implement Interface";
				}
			}
			else
			{
				if (innerClass)
				{
					return "Extend Interface with Inner Interface";
				}
				else
				{
					return "Extend Interface";
				}
			}
		}
		else if (aClass.hasModifierProperty(PsiModifier.ABSTRACT))
		{
			if (innerClass)
			{
				return "Implement Abstract Class with Inner Class";
			}
			else
			{
				return "Implement Abstract Class";
			}
		}
		else
		{
			if (innerClass)
			{
				return "Extend Class with Inner Class";
			}
			else
			{
				return "Extend Class";
			}
		}
	}

	public void invoke(@NotNull Project project, Editor editor, PsiFile file)
			throws IncorrectOperationException
	{
		final PsiElement matchingElement = findMatchingElement(file, editor);
		if (matchingElement == null)
		{
			return;
		}
		final PsiElement element = matchingElement.getParent();
		final PsiManager manager = element.getManager();
		final PsiElementFactory factory = manager.getElementFactory();
		final PsiClassType superClassType;
		if (element instanceof PsiClass)
		{
			final PsiClass superClass = (PsiClass)element;
			superClassType = factory.createType(superClass);
		}
		else
		{
			final PsiJavaCodeReferenceElement reference = (PsiJavaCodeReferenceElement)element;
			final PsiClass superClass = (PsiClass)reference.resolve();
			if (superClass == null)
			{
				return;
			}
			superClassType = factory.createType(superClass);
		}
		final CommandProcessor commandProcessor = CommandProcessor.getInstance();
		final String text = getTextForElement(matchingElement);
		commandProcessor.executeCommand(project, new Runnable()
		{
			public void run()
			{
				final String className = "My" + superClassType.getClassName();
				final SubClassDialog dialog = new SubClassDialog(className, superClassType,
						matchingElement, subclassType, innerClass, false);
				dialog.show();
				if (!dialog.isOK())
				{
					return;
				}
				final String subclassName = dialog.getSubclassName();
				final String visibility = dialog.getVisibility();
				final boolean shouldMakeAbstract = dialog.shouldMakeAbstract();
				if (innerClass)
				{
					SubclassTools.createNestedSubclass(subclassName, superClassType,
							dialog.getTargetClass(), subclassType, visibility,
							dialog.shouldMakeStatic(), shouldMakeAbstract);
				}
				else
				{
					SubclassTools.createTopLevelSubclass(subclassName, superClassType,
							dialog.getTargetDirectory(), subclassType, visibility,
							shouldMakeAbstract);
				}
			}
		}, text, text);
	}


	public boolean startInWriteAction()
	{
		return false;
	}
}