/*
 * Copyright 2006-2007 Bas Leijdekkers
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.intellij.plugins.lineage.tools;

import com.intellij.codeInsight.highlighting.HighlightManager;
import com.intellij.find.FindManager;
import com.intellij.find.FindModel;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.editor.colors.EditorColors;
import com.intellij.openapi.editor.colors.EditorColorsManager;
import com.intellij.openapi.editor.colors.EditorColorsScheme;
import com.intellij.openapi.editor.ex.MarkupModelEx;
import com.intellij.openapi.editor.markup.RangeHighlighter;
import com.intellij.openapi.editor.markup.TextAttributes;
import com.intellij.openapi.fileEditor.FileEditorManager;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.Key;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.psi.PsiElement;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

public class HighlightTool
{
	private static final Key<RangeHighlighter> HIGHLIGHT_USAGE_KEY =
			Key.create("lineage.highlight.usage");

	private HighlightTool()
	{
	}

	public static boolean highlightElements(
			@NotNull List<? extends PsiElement> elementCollection)
	{
		if (elementCollection.isEmpty())
		{
			return false;
		}
		final PsiElement firstElement = elementCollection.get(0);
		final Project project = firstElement.getProject();
		final FileEditorManager editorManager =
				FileEditorManager.getInstance(project);
		final Editor editor = editorManager.getSelectedTextEditor();
		if (editor == null)
		{
			return false;
		}
		final HighlightManager highlightManager =
				HighlightManager.getInstance(project);
		final RangeHighlighter firstHighlighter = firstElement.getUserData(HIGHLIGHT_USAGE_KEY);
		if (firstHighlighter != null) {
			final MarkupModelEx markupModel = (MarkupModelEx)editor.getMarkupModel();
			boolean highlightingRemoved = false;
			for (PsiElement element : elementCollection)
			{
				final RangeHighlighter highlighter = element.getUserData(HIGHLIGHT_USAGE_KEY);
				if (highlighter != null && markupModel.containsHighlighter(highlighter))
				{
					highlightManager.removeSegmentHighlighter(editor, highlighter);
					highlightingRemoved = true;
				}
				element.putUserData(HIGHLIGHT_USAGE_KEY, null);
			}
			if (highlightingRemoved) {
				return false;
			}
		}
		final EditorColorsManager editorColorsManager =
				EditorColorsManager.getInstance();
		final EditorColorsScheme globalScheme =
				editorColorsManager.getGlobalScheme();
		final TextAttributes textattributes =
				globalScheme.getAttributes(
						EditorColors.SEARCH_RESULT_ATTRIBUTES);
		final List<RangeHighlighter> highlighters = new ArrayList();
		final PsiElement[] elements =
				elementCollection.toArray(new PsiElement[elementCollection.size()]);
		highlightManager.addOccurrenceHighlights(
				editor, elements, textattributes, false, highlighters);
		setLineTextErrorStripeTooltip(highlighters);
		for (int i = 0; i < elements.length; i++)
		{
			final PsiElement element = elements[i];
			final RangeHighlighter rangeHighlighter = highlighters.get(i);
			element.putUserData(HIGHLIGHT_USAGE_KEY, rangeHighlighter);

		}
		final FindManager findManager = FindManager.getInstance(project);
		FindModel findModel = findManager.getFindNextModel();
		if(findModel == null)
		{
			findModel = findManager.getFindInFileModel();
		}
		findModel.setSearchHighlighters(true);
		findManager.setFindWasPerformed();
		findManager.setFindNextModel(findModel);
		return true;
	}

	protected static void setLineTextErrorStripeTooltip(Collection<RangeHighlighter> highlighters)
	{
		for (RangeHighlighter highlighter : highlighters)
		{
			final Document document = highlighter.getDocument();
			final int lineNumber = document.getLineNumber(highlighter.getStartOffset());
			final String documentText = document.getText();
			final int startOffset = document.getLineStartOffset(lineNumber);
			final int endOffset = document.getLineEndOffset(lineNumber);
			final String lineText = documentText.substring(startOffset, endOffset);
			final String escapedText = StringUtil.escapeXml(lineText.trim());
			highlighter.setErrorStripeTooltip("  " + escapedText + "  ");
		}
	}
}