/*
 * Copyright 2006-2007 Bas Leijdekkers
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.intellij.plugins.lineage.tools;

import com.intellij.codeInsight.generation.OverrideImplementUtil;
import com.intellij.openapi.application.Application;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.editor.EditorFactory;
import com.intellij.openapi.fileEditor.FileDocumentManager;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.Messages;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.psi.*;
import com.intellij.psi.util.MethodSignature;
import com.intellij.util.IncorrectOperationException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Collection;

public class SubclassTools
{
	private SubclassTools()
	{}

	@Nullable
	public static PsiClass createNestedSubclass(
			@NotNull String subclassName,
			@Nullable PsiClassType superClassType,
			@NotNull PsiClass targetClass,
			@NotNull SubclassType subclassType,
			String visibility,
			boolean makeStatic,
			boolean makeAbstract)
	{
		final Application application = ApplicationManager.getApplication();
		final ClassCreator creator =
				new ClassCreator(subclassType, subclassName, superClassType, targetClass,
						visibility, makeStatic, makeAbstract);
		application.runWriteAction(creator);
		return creator.getCreatedClass();
	}

	@Nullable
	public static PsiClass createTopLevelSubclass(
			@NotNull String subclassName,
			@Nullable PsiClassType superClassType,
			@NotNull PsiDirectory targetDirectory,
			@NotNull SubclassType subclassType,
			String visibility,
			boolean makeAbstract)
	{
		final Application application = ApplicationManager.getApplication();
		final ClassCreator creator =
				new ClassCreator(subclassType, subclassName, superClassType, targetDirectory,
						visibility, makeAbstract);
		application.runWriteAction(creator);
		return creator.getCreatedClass();
	}

	private static class ClassCreator implements Runnable
	{
		private final String className;
		@Nullable private final PsiClassType superClassType;
		private final PsiClass targetClass;
		private final PsiDirectory targetDirectory;
		private final SubclassType subclassType;
		private final String visibility;
		private final boolean makeStatic;
		private final boolean makeAbstract;

		private PsiClass createdClass = null;

		ClassCreator(
				@NotNull SubclassType subclassType,
				@NotNull String className,
				@Nullable PsiClassType superClassType,
				@NotNull PsiDirectory targetDirectory,
				@Nullable String visibility,
		        boolean makeAbstract)
		{
			this.subclassType = subclassType;
			this.className = className;
			this.superClassType = superClassType;
			this.targetDirectory = targetDirectory;
			targetClass = null;
			this.visibility = visibility;
			makeStatic = false;
			this.makeAbstract = makeAbstract;
		}

		ClassCreator(
				SubclassType subclassType,
				@NotNull String className,
				@Nullable PsiClassType superClassType,
				@NotNull PsiClass targetClass,
				@Nullable String visibility,
				boolean makeStatic,
		        boolean makeAbstract)
		{
			this.subclassType = subclassType;
			this.className = className;
			this.superClassType = superClassType;
			this.targetClass = targetClass;
			targetDirectory = null;
			this.visibility = visibility;
			this.makeStatic = makeStatic;
			this.makeAbstract = makeAbstract;
		}

		private PsiClass createInnerClass() throws IncorrectOperationException
		{
			final PsiManager manager = targetClass.getManager();
			final PsiElementFactory elementFactory = manager.getElementFactory();
			PsiClass newClass;
			switch (subclassType)
			{
				case ANNOTATION_TYPE:
					newClass = elementFactory.createClassFromText("@interface " + className + "{}",
							targetClass);
					break;
				case CLASS:
					newClass = elementFactory.createClass(className);
					break;
				case ENUM:
					newClass = elementFactory.createEnum(className);
					break;
				case INTERFACE:
					newClass = elementFactory.createInterface(className);
					break;
				default:
					throw new InternalError();
			}
			newClass = (PsiClass)targetClass.add(newClass);
			if (superClassType != null)
			{
				final PsiClassType javaLangObject =
						PsiType.getJavaLangObject(manager, superClassType.getResolveScope());
				if (!superClassType.equals(javaLangObject))
				{
					final PsiClass superClass;
					if (superClassType != null)
					{
						superClass = superClassType.resolve();
					}
					else
					{
						superClass =  null;
					}
					final PsiJavaCodeReferenceElement classReference =
							elementFactory.createReferenceElementByType(superClassType);
					final PsiReferenceList referenceList;
					if (superClass != null && superClass.isInterface() &&
							subclassType != SubclassType.INTERFACE)
					{
						referenceList = newClass.getImplementsList();
					}
					else if (superClass != newClass)
					{
						referenceList = newClass.getExtendsList();
					}
					else
					{
						referenceList = null;
					}
					if (referenceList != null)
					{
						referenceList.add(classReference);
					}
				}
			}
			if (makeStatic)
			{
				final PsiModifierList modifierList = newClass.getModifierList();
				if (modifierList != null)
				{
					modifierList.setModifierProperty(PsiModifier.STATIC, true);
				}
			}
			newClass.navigate(true);
			return newClass;
		}

		private PsiClass createToplevelClass()
				throws IncorrectOperationException
		{
			final PsiClass newClass;
			switch (subclassType)
			{

				case ANNOTATION_TYPE:
					newClass = targetDirectory.createAnnotationType(className);
					break;
				case CLASS:
					newClass = targetDirectory.createClass(className);
					break;
				case ENUM:
					newClass = targetDirectory.createEnum(className);
					break;
				case INTERFACE:
					newClass = targetDirectory.createInterface(className);
					break;
				default:
					throw new InternalError("unknown subclass type");
			}
			if (superClassType != null)
			{
				final PsiManager manager = targetDirectory.getManager();
				final PsiElementFactory elementFactory = manager.getElementFactory();
				final PsiClassType javaLangObject =
						PsiType.getJavaLangObject(manager, superClassType.getResolveScope());
				if (!superClassType.equals(javaLangObject))
				{
					final PsiJavaCodeReferenceElement classReference =
							elementFactory.createReferenceElementByType(superClassType);
					final PsiReferenceList referenceList;
					final PsiClass superClass = superClassType.resolve();
					if (subclassType != SubclassType.INTERFACE && superClass != null &&
							superClass.isInterface())
					{
						referenceList = newClass.getImplementsList();
					}
					else
					{
						if (superClass != newClass)
						{
							referenceList = newClass.getExtendsList();
						}
						else
						{
							referenceList = null;
						}
					}
					if (referenceList != null)
					{
						referenceList.add(classReference);
					}
				}
			}
			newClass.navigate(true);
			return newClass;
		}

		@Nullable
		public PsiClass getCreatedClass()
		{
			return createdClass;
		}

		public void run()
		{
			final Project project;
			if (targetClass == null)
			{
				project = targetDirectory.getProject();
			}
			else
			{
				project = targetClass.getProject();
			}
			try
			{
				final PsiClass newClass;
				if (targetClass == null)
				{
					newClass = createToplevelClass();
				}
				else
				{
					newClass = createInnerClass();
				}
				final PsiModifierList modifierList = newClass.getModifierList();
				if (visibility != null && modifierList != null)
				{
					modifierList.setModifierProperty(visibility, true);
				}
				if (makeAbstract && modifierList != null)
				{
					modifierList.setModifierProperty(PsiModifier.ABSTRACT, true);
				}
				final PsiFile containingFile = newClass.getContainingFile();
				final VirtualFile virtualFile = containingFile.getVirtualFile();
				final FileDocumentManager fileDocumentManager =
						FileDocumentManager.getInstance();
				final Document document;
				if (virtualFile != null)
				{
					document = fileDocumentManager.getDocument(virtualFile);
				}
				else
				{
					document = null;
				}
				final EditorFactory editorFactory = EditorFactory.getInstance();
				final Editor currentEditor;
				if (document != null)
				{
					currentEditor = editorFactory.getEditors(document)[0];
				}
				else
				{
					currentEditor = null;
				}
				final SmartPointerManager pointerManager =
						SmartPointerManager.getInstance(project);
				final SmartPsiElementPointer<PsiClass> pointer =
						pointerManager.createSmartPsiElementPointer(newClass);
				if (superClassType != null)
				{
					final PsiClass superClass = superClassType.resolve();
					// todo create combined implement/override dialog.
					if (superClass != null && (!superClass.isInterface() ||
							superClass.hasModifierProperty(
									PsiModifier.ABSTRACT)) &&
							subclassType != SubclassType.INTERFACE &&
							subclassType != SubclassType.ANNOTATION_TYPE)
					{
						OverrideImplementUtil.chooseAndImplementMethods(project, currentEditor,
								newClass);
						final Collection<MethodSignature> signatures =
								OverrideImplementUtil.getMethodSignaturesToImplement(newClass);
						if (!signatures.isEmpty() && modifierList != null)
						{
							modifierList.setModifierProperty(PsiModifier.ABSTRACT, true);
						}
					}
					else
					{
						final PsiManager manager = newClass.getManager();
						final PsiClassType javaLangObject =
								PsiType.getJavaLangObject(manager,
										superClassType.getResolveScope());
						if (!superClassType.equals(javaLangObject)) {
							OverrideImplementUtil.chooseAndOverrideMethods(project, currentEditor,
									pointer.getElement());
						}
					}
				}
				createdClass = pointer.getElement();
			}
			catch (IncorrectOperationException ignore)
			{
				Messages.showErrorDialog(project, "Cannot Create Class '" + className + '\'',
						"Failed to Create Class");
			}
		}
	}
}