/*
 * Copyright 2006-2007 Sascha Weinreuter, Bas Leijdekkers
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.intellij.plugins.lineage.ui;

import com.intellij.openapi.actionSystem.AnAction;
import com.intellij.openapi.actionSystem.CustomShortcutSet;
import com.intellij.openapi.actionSystem.AnActionEvent;
import com.intellij.ui.EditorTextField;
import com.intellij.ui.EditorComboBox;

import javax.swing.*;
import java.awt.event.KeyEvent;
import java.awt.event.InputEvent;
import java.awt.Container;
import java.awt.FocusTraversalPolicy;
import java.awt.Component;

/**
 * Provides Shift-Tab support in EditorTextFields which otherwise don't support this keystroke to
 * move the input focus to the previous component.
 */
@SuppressWarnings({"ComponentNotRegistered"})
class ShiftTabAction extends AnAction
{
	private static final CustomShortcutSet SHIFT_TAB;

	static
	{
		final KeyStroke keyStroke = KeyStroke.getKeyStroke(KeyEvent.VK_TAB, InputEvent.SHIFT_MASK);
		SHIFT_TAB = new CustomShortcutSet(keyStroke);
	}

	private final JComponent component;

	private ShiftTabAction(JComponent component)
	{
		super("Shift-Tab");
		this.component = component;
		setInjectedContext(isInInjectedContext());
	}

	public void actionPerformed(AnActionEvent event)
	{
		Container container = component.getParent();
		while (container != null && container.getFocusTraversalPolicy() == null)
		{
			container = container.getParent();
		}
		if (container != null)
		{
			final FocusTraversalPolicy ftp = container.getFocusTraversalPolicy();
			if (ftp != null)
			{
				final Component prev = ftp.getComponentBefore(container, component);
				if (prev != null)
				{
					prev.requestFocus();
				}
			}
		}
	}

	/**
	 * Enables Sift-Tab support for the supplied EditorTextField.
	 */
	public static void attachTo(EditorTextField textField)
	{
		final ShiftTabAction action = new ShiftTabAction(textField);
		action.registerCustomShortcutSet(SHIFT_TAB, textField);
	}

	public static void attachTo(EditorComboBox editorComboBox)
	{
		final ShiftTabAction action = new ShiftTabAction(editorComboBox);
		action.registerCustomShortcutSet(SHIFT_TAB, editorComboBox);
	}
}