/*
 * Copyright 2006-2007 Bas Leijdekkers
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.intellij.plugins.lineage.ui;

import com.intellij.ide.util.PackageChooserDialog;
import com.intellij.ide.util.PackageUtil;
import com.intellij.ide.util.TreeClassChooser;
import com.intellij.ide.util.TreeClassChooserFactory;
import com.intellij.lang.StdLanguages;
import com.intellij.lang.refactoring.NamesValidator;
import com.intellij.openapi.editor.event.DocumentAdapter;
import com.intellij.openapi.editor.event.DocumentEvent;
import com.intellij.openapi.fileTypes.StdFileTypes;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.*;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.psi.*;
import com.intellij.psi.search.GlobalSearchScope;
import com.intellij.psi.util.ClassUtil;
import com.intellij.psi.util.InheritanceUtil;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.refactoring.ui.ConflictsDialog;
import com.intellij.refactoring.util.ConflictsUtil;
import com.intellij.refactoring.util.RefactoringMessageUtil;
import com.intellij.refactoring.util.VisibilityUtil;
import com.intellij.ui.EditorComboBox;
import com.intellij.ui.EditorTextField;
import com.intellij.ui.IdeBorderFactory;
import com.intellij.ui.ReferenceEditorComboWithBrowseButton;
import com.intellij.util.IncorrectOperationException;
import net.intellij.plugins.lineage.tools.SubclassType;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import javax.swing.border.Border;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;

public class SubClassDialog extends DialogWrapper
{
	private final JLabel extendLabel = new JLabel("Extend");
	private final LabeledComponent<EditorTextField> nameComponent = new LabeledComponent();
	private final JCheckBox declareStaticCheckBox = new JCheckBox("Declare static");
	private final JCheckBox declareAbstractCheckBox = new JCheckBox("Declare abstract");
	private final LabeledComponent<ReferenceEditorComboWithBrowseButton> destinationComponent =
			new LabeledComponent();
	private final Border titledBorder;
	private final ButtonGroup visibilityButtonGroup;
	private final Project project;
	private final PsiElement context;
	private final SubclassType type;
	private final boolean createInnerClass;
	private final boolean createNewClass;
	@Nullable private final PsiClass superClass;

	private PsiClass targetClass = null;
	PsiDirectory targetDirectory = null;

	private boolean checkVisibility = false;

	public SubClassDialog(@NotNull String subclassName,
	                      @Nullable PsiClassType superClassType,
	                      @NotNull PsiElement context,
	                      SubclassType type,
	                      boolean createInnerClass,
	                      boolean createNewClass)
	{
		super(context.getProject(), true);
		this.context = context;
		project = context.getProject();
		this.type = type;
		this.createInnerClass = createInnerClass;
		this.createNewClass = createNewClass;
		visibilityButtonGroup = createVisibilityButtonGroup();
		final PsiClass target;
		if (superClassType != null)
		{
			target = superClassType.resolve();
		}
		else
		{
			target = null;
		}
		if (target != null)
		{
			superClass = target;
		}
		else
		{
			final PsiManager manager = context.getManager();
			superClass = manager.findClass("java.lang.Object", context.getResolveScope());
		}
		if (!createNewClass)
		{
			final boolean isInterface =
					superClass != null && superClass.isInterface();
			if (type == SubclassType.INTERFACE ||
					!isInterface)
			{
				setOKButtonText("&Extend");
			}
			else
			{
				setOKButtonText("&Implement");
			}
			final String superclassName;
			if (superClassType != null)
			{
				superclassName = superClassType.getCanonicalText();
			}
			else
			{
				superclassName = CommonClassNames.JAVA_LANG_OBJECT;
			}
			if (isInterface)
			{
				if (type == SubclassType.INTERFACE)
				{
					setTitle("Extend Interface");
					extendLabel.setText("Extend interface '" +
							superclassName + "\' with");
					titledBorder = IdeBorderFactory.createTitledBorder("Interface");
				}
				else
				{
					setTitle("Implement Interface");
					extendLabel.setText("Implement interface '" +
							superclassName + "\' with");
					titledBorder = IdeBorderFactory.createTitledBorder("Class");
				}
			}
			else
			{
				setTitle("Extend Class");
				extendLabel.setText("Extend class '" + superclassName + "\' with");
				titledBorder = IdeBorderFactory.createTitledBorder("Class");
			}
		}
		else
		{
			setOKButtonText("&Create");
			final String typeString = StringUtil.capitalize(type.toString());
			setTitle("New " + typeString);
			titledBorder = IdeBorderFactory.createTitledBorder(typeString);
		}
		init(subclassName, context, createInnerClass);
	}


	private void checkInnerClassCreation(@NotNull String subclassName,
	                                     @NotNull Collection<String> problems)
	{
		if (subclassName.equals(targetClass.getName()))
		{
			final String description = ConflictsUtil.getDescription(targetClass, true);
			problems.add(description + "</b> can't contain an inner " + type +
					" named <b>" + subclassName + "</b>");
			return;
		}
		final PsiClass[] innerClasses = targetClass.getInnerClasses();
		for (PsiClass innerClass : innerClasses)
		{
			if (subclassName.equals(innerClass.getName()))
			{
				final String description1 = ConflictsUtil.getDescription(targetClass, true);
				final String description2 = ConflictsUtil.getDescription(innerClass, true);
				problems.add(description1 + " already contains a inner " + description2);
			}
		}
		final PsiMember member = PsiTreeUtil.getParentOfType(context, PsiMember.class);
		final boolean declareStatic = declareStaticCheckBox.isSelected();
		if (member != null && member.hasModifierProperty(PsiModifier.STATIC) && !declareStatic)
		{
			problems.add("Non-static inner " + type + " <b>" + subclassName +
					"</b> will not be accessible from a static context");
		}
		final PsiClass containingClass = PsiTreeUtil.getParentOfType(context, PsiClass.class);
		// todo might target class be null?
		if (!targetClass.equals(containingClass) && !declareStatic)
		{
			final String description = ConflictsUtil.getDescription(targetClass, true);
			problems.add("Non-static inner " + type + " <b>" + subclassName +
					"</b> will not be accessible without an instance of " + description);
		}
	}

	protected JComponent createCenterPanel()
	{
		final ListFocusTraversalPolicy focusTraversalPolicy = new ListFocusTraversalPolicy();

		final GridBagConstraints constraints = new GridBagConstraints();
		constraints.anchor = GridBagConstraints.NORTHWEST;
		constraints.fill = GridBagConstraints.HORIZONTAL;
		constraints.gridwidth = 2;
		constraints.insets.left = 4;
		constraints.weightx = 1.0;
		final JPanel centerPanel = new JPanel(new GridBagLayout());
		if (!createNewClass) {
			centerPanel.add(extendLabel, constraints);
		}

		constraints.gridwidth = 1;
		constraints.gridy = 1;
		constraints.insets.top = 4;
		constraints.insets.left = 0;
		constraints.weighty = 1.0;
		constraints.fill = GridBagConstraints.BOTH;
		final JComponent namePanel = createNamePanel(focusTraversalPolicy);
		centerPanel.add(namePanel, constraints);

		constraints.gridx = 0;
		constraints.weighty = 0.0;
		constraints.gridy = 2;
		constraints.insets.left = 4;
		constraints.gridwidth = 2;
		final ReferenceEditorComboWithBrowseButton comboWithBrowseButton =
				destinationComponent.getComponent();
		final EditorComboBox editorComboBox = comboWithBrowseButton.getChildComponent();
		focusTraversalPolicy.add(editorComboBox);
		centerPanel.add(destinationComponent, constraints);

		constraints.gridx = 1;
		constraints.gridy = 1;
		constraints.weightx = 0.0;
		constraints.weighty = 1.0;
		constraints.gridwidth = 1;
		centerPanel.add(createVisibilityPanel(focusTraversalPolicy), constraints);

		centerPanel.setFocusTraversalPolicy(focusTraversalPolicy);
		centerPanel.setFocusTraversalPolicyProvider(true);
		//centerPanel.setFocusCycleRoot(true);
		return centerPanel;
	}

	private JComponent createNamePanel(
			ListFocusTraversalPolicy focusTraversalPolicy)
	{
		final JPanel namePanel = new JPanel(new GridBagLayout());
		namePanel.setBorder(titledBorder);
		final GridBagConstraints constraints = new GridBagConstraints();
		constraints.weightx = 1.0;
		constraints.gridx = 0;
		constraints.gridy = 0;
		if (createInnerClass)
		{
			constraints.weighty = 0.0;
		}
		else
		{
			constraints.weighty = 1.0;
		}
		constraints.anchor = GridBagConstraints.NORTHWEST;
		constraints.fill = GridBagConstraints.HORIZONTAL;
		constraints.insets = new Insets(4, 4, 4, 4);
		namePanel.add(nameComponent, constraints);
		final EditorTextField editorTextField = nameComponent.getComponent();
		focusTraversalPolicy.add(editorTextField);
		if (createInnerClass)
		{
			if (type == SubclassType.CLASS)
			{
				constraints.weighty = 0.0;
			}
			else
			{
				constraints.weighty = 1.0;
			}
			constraints.gridy++;
			constraints.fill = GridBagConstraints.NONE;
			constraints.insets.top = 0;
			constraints.insets.bottom = 0;
			namePanel.add(declareStaticCheckBox, constraints);
			focusTraversalPolicy.add(declareStaticCheckBox);
		}
		if (type == SubclassType.CLASS)
		{
			constraints.weighty = 1.0;
			constraints.gridy = constraints.gridy + 1;
			constraints.fill = GridBagConstraints.NONE;
			constraints.insets.top = 0;
			constraints.insets.bottom = 0;
			namePanel.add(declareAbstractCheckBox, constraints);
			focusTraversalPolicy.add(declareAbstractCheckBox);
		}
		return namePanel;
	}

	@Nullable
	protected JComponent createSouthPanel()
	{
		final JComponent superPanel = super.createSouthPanel();
		final JLabel poweredLabel = new JLabel("Powered by Lineage");
		poweredLabel.setEnabled(false);
		final Font font = poweredLabel.getFont();
		final float size = (float)font.getSize();
		poweredLabel.setFont(font.deriveFont(size - 2.0f));
		final Insets insets = poweredLabel.getInsets();
		insets.top += 25;
		insets.left += 4;
		poweredLabel.setBorder(IdeBorderFactory.createEmptyBorder(insets));

		final JComponent wrappingPanel = new JPanel(new BorderLayout());
		wrappingPanel.add(poweredLabel, BorderLayout.WEST);
		if (superPanel != null)
		{
			wrappingPanel.add(superPanel, BorderLayout.EAST);
		}
		return wrappingPanel;
	}

	private ButtonGroup createVisibilityButtonGroup()
	{
		final ButtonGroup visibilityButtonGroup = new ButtonGroup();
		if (createInnerClass)
		{
			final JRadioButton privateRadioButton = new JRadioButton("Private");
			privateRadioButton.setMnemonic('v');
			privateRadioButton.setActionCommand(PsiModifier.PRIVATE);
			if (type != SubclassType.INTERFACE)
			{
				privateRadioButton.setSelected(true);
			}
			visibilityButtonGroup.add(privateRadioButton);
		}
		final JRadioButton packageRadioButton = new JRadioButton("Package local");
		packageRadioButton.setMnemonic('k');
		packageRadioButton.setActionCommand(PsiModifier.PACKAGE_LOCAL);
		visibilityButtonGroup.add(packageRadioButton);
		if (createInnerClass)
		{
			final JRadioButton protectedRadioButton = new JRadioButton("Protected");
			protectedRadioButton.setMnemonic('o');
			protectedRadioButton.setActionCommand(PsiModifier.PROTECTED);
			visibilityButtonGroup.add(protectedRadioButton);
		}
		final JRadioButton publicRadioButton = new JRadioButton("Public");
		publicRadioButton.setMnemonic('b');
		if (!createInnerClass)
		{
			publicRadioButton.setSelected(true);
		}
		else if (type == SubclassType.INTERFACE)
		{
			publicRadioButton.setSelected(true);
		}
		publicRadioButton.setActionCommand(PsiModifier.PUBLIC);
		visibilityButtonGroup.add(publicRadioButton);
		return visibilityButtonGroup;
	}

	private JComponent createVisibilityPanel(
			ListFocusTraversalPolicy focusTraversalPolicy)
	{
		final JPanel visibilityPanel =
				new JPanel(new VerticalFlowLayout(FlowLayout.LEFT, 0, 0, true, false));
		visibilityPanel.setBorder(IdeBorderFactory.createTitledBorder("Visibility"));
		final Enumeration<AbstractButton> enumeration = visibilityButtonGroup.getElements();
		while (enumeration.hasMoreElements())
		{
			final AbstractButton radioButton = enumeration.nextElement();
			visibilityPanel.add(radioButton);
			focusTraversalPolicy.add(radioButton);
			if (type == SubclassType.INTERFACE && createInnerClass)
			{
				radioButton.setEnabled(false);
			}
		}
		return visibilityPanel;
	}

	protected void doOKAction()
	{
		final String targetName = getTargetName();
		final String subclassName = getSubclassName();
		final Collection<String> problems = new ArrayList();
		if (createInnerClass)
		{
			targetClass = findOrCreateClass(targetName);
			if (targetClass == null)
			{
				return;
			}
			checkInnerClassCreation(subclassName, problems);
			if (checkVisibility && !isVisible(getVisibility(), targetClass, context))
			{
				final PsiMember member = PsiTreeUtil.getParentOfType(context, PsiMember.class);
				if (member != null)
				{
					final String memberDescription = ConflictsUtil.getDescription(member, true);
					final String problemString = type + " <b>" + targetName + '.' +
							subclassName + "</b> is " + getVisibility() +
							" and will not be accessible from " + memberDescription;
					problems.add(problemString);
				}
			}
			if (superClass != null)
			{
				final PsiModifierList modifierList = superClass.getModifierList();
				final String visibility = VisibilityUtil.getVisibilityModifier(modifierList);
				if (!isVisible(visibility, superClass, targetClass))
				{
					final String problemString = "Super " +
							ConflictsUtil.getDescription(superClass, true) + " is " +
							visibility + " and will not be accessible from " +
							ConflictsUtil.getDescription(targetClass, true);
					problems.add(problemString);
				}
			}
		}
		else
		{
			try
			{
				final PsiDirectory tmpDirectory;
				if (context instanceof PsiDirectory)
				{
					tmpDirectory = (PsiDirectory)context;
				}
				else
				{
					tmpDirectory = PsiTreeUtil.getParentOfType(context, PsiDirectory.class);
				}
				targetDirectory =
						PackageUtil.findOrCreateDirectoryForPackage(project, targetName,
								tmpDirectory, true);
				if (targetDirectory == null)
				{
					return;
				}
				final String createClassProblem = RefactoringMessageUtil.checkCanCreateClass(
						targetDirectory, subclassName);
				if (createClassProblem != null && createClassProblem.length() > 0) {
					Messages.showMessageDialog(project, createClassProblem, "Error",
							Messages.getErrorIcon());
					return;
				}
				if (superClass != null)
				{
					final PsiModifierList modifierList = superClass.getModifierList();
					final String visibility = VisibilityUtil.getVisibilityModifier(modifierList);
					final PsiPackage targetPackage = targetDirectory.getPackage();
					if (targetPackage != null && !isVisible(visibility, superClass, targetPackage))
					{
						final String description;
						if (targetPackage.getName() == null)
						{
							description = "<b><code>&lt;Default Package&gt;</code></b>";
						}
						else
						{
							description = ConflictsUtil.getDescription(targetPackage, true);
						}
						final String problemString = "Super " +
								ConflictsUtil.getDescription(superClass, true) + " is " +
								visibility + " and will not be accessible from " + description;
						problems.add(problemString);
					}
				}
			}
			catch (IncorrectOperationException incorrectoperationexception)
			{
				final String incorrectOperationError = incorrectoperationexception.getMessage();
				Messages.showMessageDialog(project, incorrectOperationError, "Error",
						Messages.getErrorIcon());
				return;
			}
		}
		if (!problems.isEmpty())
		{
			final ConflictsDialog conflictsDialog = new ConflictsDialog(project, problems);
			conflictsDialog.show();
			if (!conflictsDialog.isOK())
			{
				return;
			}
		}
		super.doOKAction();
	}

	@Nullable
	private PsiClass findOrCreateClass(String targetName)
	{
		final PsiManager manager = PsiManager.getInstance(project);
		final PsiClass result = manager.findClass(targetName, GlobalSearchScope.allScope(project));
		if (result != null)
		{
			return result;
		}
		final int yesNo = Messages.showYesNoDialog(project,
				"Class " + targetName + " does not exist.\nDo you wish to create it?",
				"Class not found", Messages.getQuestionIcon());
		if (yesNo != 0)
		{
			return null;
		}
		final String packageName = ClassUtil.extractPackageName(targetName);
		try
		{
			final PsiDirectory directory =
					PackageUtil.findOrCreateDirectoryForPackage(project, packageName,
							null, true);
			if (directory == null)
			{
				return null;
			}
			final String targetClassName = ClassUtil.extractClassName(targetName);
			return directory.createClass(targetClassName);
		}
		catch (IncorrectOperationException e)
		{
			Messages.showMessageDialog(project, e.getMessage(), "Error",
					Messages.getErrorIcon());
			return null;
		}
	}

	public JComponent getPreferredFocusedComponent()
	{
		return nameComponent.getComponent();
	}

	public String getSubclassName()
	{
		final EditorTextField nameTextField = nameComponent.getComponent();
		return nameTextField.getText();
	}

	public PsiClass getTargetClass()
	{
		return targetClass;
	}

	public PsiDirectory getTargetDirectory()
	{
		return targetDirectory;
	}

	private String getTargetName()
	{
		final ReferenceEditorComboWithBrowseButton destinationTextField =
				destinationComponent.getComponent();
		final String destinationName = destinationTextField.getText();
		if (destinationName == null)
		{
			return "";
		}
		else
		{
			return destinationName.trim();
		}
	}

	public String getVisibility()
	{
		final ButtonModel model = visibilityButtonGroup.getSelection();
		return model.getActionCommand();
	}

	private void init(String subclassName, PsiElement context, boolean createInnerClass)
	{
		final PsiPackage containingPackage;
		PsiClass containingClass;
		if (createInnerClass)
		{
			containingClass = PsiTreeUtil.getParentOfType(context, PsiClass.class);
			while (containingClass instanceof PsiAnonymousClass)
			{
				containingClass = PsiTreeUtil.getParentOfType(containingClass, PsiClass.class);
			}
			containingPackage = null;
		}
		else
		{
			final PsiDirectory directory;
			if (context instanceof PsiDirectory)
			{
				directory = (PsiDirectory) context;
			}
			else
			{
				final PsiFile file = context.getContainingFile();
				directory = file.getContainingDirectory();
			}
			assert directory != null;
			containingPackage = directory.getPackage();
			containingClass = null;
		}
		final Project project = context.getProject();
		final EditorTextField nameTextField =
				new EditorTextField(subclassName, project, StdFileTypes.JAVA);
		nameTextField.addDocumentListener(new DocumentAdapter()
		{
			public void documentChanged(DocumentEvent event)
			{
				final NamesValidator namesValidator = StdLanguages.JAVA.getNamesValidator();
				final boolean validIdentifier =
						namesValidator.isIdentifier(nameTextField.getText(), project);
				setOKActionEnabled(validIdentifier);
			}
		});
		nameTextField.selectAll();
		final Dimension preferredSize = nameTextField.getPreferredSize();
		preferredSize.width = 200;
		nameTextField.setMinimumSize(preferredSize);
		nameComponent.setComponent(nameTextField);
		nameComponent.setText("&Name");
		declareStaticCheckBox.setMnemonic('s');
		declareAbstractCheckBox.setMnemonic('a');
		if (createInnerClass)
		{
			final PsiMember member = PsiTreeUtil.getParentOfType(context,
					PsiMember.class);
			if (member != null && member.hasModifierProperty(PsiModifier.STATIC))
			{
				declareStaticCheckBox.setSelected(true);
			}
			initClassChooser(containingClass);
		}
		else
		{
			initPackageChooser(containingPackage);
		}
		init();
	}

	private void initClassChooser(final PsiClass initialClass)
	{
		final Project project = initialClass.getProject();
		final ReferenceEditorComboWithBrowseButton destinationField =
				new ReferenceEditorComboWithBrowseButton(new ActionListener()
				{
					public void actionPerformed(ActionEvent actionEvent)
					{
						final TreeClassChooserFactory factory =
								TreeClassChooserFactory.getInstance(project);
						final TreeClassChooser chooser =
								factory.createWithInnerClassesScopeChooser(
										"Choose Destination Class",
										GlobalSearchScope.projectScope(project),
										null, initialClass);
						chooser.showDialog();
						final PsiClass selectedClass = chooser.getSelectedClass();
						if (selectedClass != null)
						{
							final ReferenceEditorComboWithBrowseButton destinationField =
									destinationComponent.getComponent();
							destinationField.prependItem(selectedClass.getQualifiedName());
						}
					}
				}, initialClass.getQualifiedName(), initialClass.getManager(), true,
						"SubclassDialog.RECENT_CLASSES_KEY");
		//destinationField.getChildComponent().addDocumentListener(new DocumentAdapter()
		//{
		//	public void documentChanged(DocumentEvent event)
		//	{
		//		final PsiManager manager = PsiManager.getInstance(project);
		//		final String targetName = getTargetName();
		//		final PsiClass result = manager.findClass(targetName,
		//				GlobalSearchScope.allScope(project));
		//		if (result != null)
		//		{
		//			final PsiClass contextClass =
		//					PsiTreeUtil.getParentOfType(context, PsiClass.class);
		//			if (!result.equals(contextClass)) {
		//				final PsiPackage resultPackage =
		//						PsiTreeUtil.getParentOfType(result, PsiPackage.class);
		//				final PsiPackage contextPackage =
		//						PsiTreeUtil.getParentOfType(contextClass, PsiPackage.class);
		//				if (resultPackage.equals(contextPackage)) {
		//
		//				}
		//			}
		//		}
		//	}
		//});
		final EditorComboBox editorComboBox = destinationField.getChildComponent();
		ShiftTabAction.attachTo(editorComboBox);
		destinationComponent.setComponent(destinationField);
		destinationComponent.setText("&Destination Class");
	}

	private void initPackageChooser(PsiPackage initialPackage)
	{
		final Project project = initialPackage.getProject();
		final PsiManager manager = initialPackage.getManager();
		final ReferenceEditorComboWithBrowseButton destinationField =
				new ReferenceEditorComboWithBrowseButton(
						new ActionListener()
						{
							public void actionPerformed(ActionEvent actionEvent)
							{
								final ReferenceEditorComboWithBrowseButton destinationField =
										destinationComponent.getComponent();
								final PackageChooser packageDialog =
										new PackageChooserDialog("Choose Destination Package",
												project);
								packageDialog.selectPackage(destinationField.getText());
								packageDialog.show();
								final PsiPackage aPackage = packageDialog.getSelectedPackage();
								if (aPackage != null)
								{
									destinationField.prependItem(aPackage.getQualifiedName());
								}
							}
						}, initialPackage.getQualifiedName(), manager, false,
						"SubClassDialog.RECENT_PACKAGES_KEY");
		destinationComponent.setComponent(destinationField);
		destinationComponent.setText("&Destination Package");
		final EditorComboBox childComponent = destinationField.getChildComponent();
		childComponent.addDocumentListener(new DocumentAdapter()
		{
			public void documentChanged(DocumentEvent e)
			{
				final PsiNameHelper nameHelper = manager.getNameHelper();
				final String packageText = destinationField.getText();
				final boolean validPackageName = packageText.length() == 0 ||
						nameHelper.isQualifiedName(packageText);
				setOKActionEnabled(validPackageName);
			}
		});
	}

	@SuppressWarnings({"StringEquality"})
	private static boolean isVisible(@NotNull String visibility,
	                                 @NotNull PsiClass superClass,
	                                 @NotNull PsiElement accessLocation)
	{
		if (PsiModifier.PRIVATE == visibility)
		{
			final PsiClass parentClass =
					PsiTreeUtil.getParentOfType(accessLocation, PsiClass.class, false);
			if (superClass == parentClass)
			{
				return true;
			}
			final PsiClass containingClass = superClass.getContainingClass();
			return containingClass == parentClass;
		}
		else if (PsiModifier.PACKAGE_LOCAL == visibility)
		{
			final PsiFile containingFile = superClass.getContainingFile();
			if (!(containingFile instanceof PsiJavaFile))
			{
				return false;
			}
			final PsiJavaFile javaFile = (PsiJavaFile)containingFile;
			final String superPackageName = javaFile.getPackageName();
			final PsiPackage parentPackage =
					PsiTreeUtil.getParentOfType(accessLocation, PsiPackage.class, false);
			final String parentPackageName;
			if (parentPackage == null)
			{
				parentPackageName = "";
			}
			else
			{
				parentPackageName = parentPackage.getQualifiedName();
			}
			return superPackageName.equals(parentPackageName);
		}
		else if (PsiModifier.PROTECTED == visibility)
		{
			final PsiFile containingFile = superClass.getContainingFile();
			if (!(containingFile instanceof PsiJavaFile))
			{
				return false;
			}
			final PsiJavaFile javaFile = (PsiJavaFile)containingFile;
			final String superPackageName = javaFile.getPackageName();
			final PsiClass parentClass =
					PsiTreeUtil.getParentOfType(accessLocation, PsiClass.class, false);
			if (InheritanceUtil.isCorrectDescendant(parentClass, superClass, true))
			{
				return true;
			}
			final PsiPackage parentPackage =
					PsiTreeUtil.getParentOfType(parentClass, PsiPackage.class);
			final String parentPackageName;
			if (parentPackage == null)
			{
				parentPackageName = "";
			}
			else
			{
				parentPackageName = parentPackage.getQualifiedName();
			}
			return superPackageName.equals(parentPackageName);
		}
		else if (PsiModifier.PUBLIC == visibility)
		{
			return true;
		}
		throw new IllegalArgumentException("unknown modifier: " + visibility);
	}

	private static void printConstraints(GridBagConstraints constraints) {
		final StringBuilder builder = new StringBuilder("GridBagConstraints[");
		builder.append("gridx=");
		builder.append(constraints.gridx);
		builder.append(", gridy=");
		builder.append(constraints.gridy);
		builder.append(", gridwidth=");
		builder.append(constraints.gridwidth);
		builder.append(", gridheight=");
		builder.append(constraints.gridheight);
		builder.append(", weightx=");
		builder.append(constraints.weightx);
		builder.append(", weighty=");
		builder.append(constraints.weighty);
		System.out.println(builder.toString());
	}

	public void setCheckVisibilityFromContext(boolean checkVisibility)
	{
		this.checkVisibility = checkVisibility;
	}

	public boolean shouldCreateInnerClass()
	{
		return createInnerClass;
	}

	public boolean shouldMakeAbstract()
	{
		return declareAbstractCheckBox.isSelected();
	}

	public boolean shouldMakeStatic()
	{
		return declareStaticCheckBox.isSelected();
	}
}